/* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */

const Clutter = imports.gi.Clutter;
const Pango = imports.gi.Pango;
const GLib = imports.gi.GLib;
const Gio = imports.gi.Gio;
const Gtk = imports.gi.Gtk;
const Shell = imports.gi.Shell;
const Lang = imports.lang;
const Signals = imports.signals;
const St = imports.gi.St;
const Mainloop = imports.mainloop;

const AppFavorites = imports.ui.appFavorites;
const DND = imports.ui.dnd;
const Main = imports.ui.main;
const Overview = imports.ui.overview;
const PopupMenu = imports.ui.popupMenu;
const Search = imports.ui.search;
const Tweener = imports.ui.tweener;
const Workspace = imports.ui.workspace;
const AppDisplay = imports.ui.appDisplay;
const AltTab = imports.ui.altTab;

const Dash = imports.ui.dash;
const Meta = imports.gi.Meta;
const IconGrid = imports.ui.iconGrid;


//const Gettext = imports.gettext.domain('gnome-shell-extensions');
//const _ = Gettext.gettext;

// Settings
const DOCK_SETTINGS_SCHEMA = 'org.gnome.shell.extensions.dock';
const DOCK_SIZE_KEY = 'size';
const DOCK_HIDE_KEY = 'autohide';
const DOCK_AUTOHIDE_ANIMATION_TIME_KEY = 'hide-effect-duration';

//hide
//const autohide_animation_time = 0.3;

// Keep enums in sync with GSettings schemas

let dockicon_size = 48;
let hideable = true;
let hideDock = true;
let autohide_animation_time = 0.3;
const DND_RAISE_APP_TIMEOUT = 500;

const DASH_ANIMATION_TIME = Dash.DASH_ANIMATION_TIME;

const DashItemContainer = Dash.DashItemContainer;

const RemoveFavoriteIcon = Dash.RemoveFavoriteIcon;

const DragPlaceholderItem = Dash.DragPlaceholderItem;

const OrigDash = Dash.Dash;

function BottomDash() {
    this._init();
}

BottomDash.prototype = {
    __proto__: OrigDash.prototype,
    _init : function() {
        this._maxWidth = -1;
        this._shownInitially = false;

        this._dragPlaceholder = null;
        this._dragPlaceholderPos = -1;
        this._animatingPlaceholdersCount = 0;
        this._favRemoveTarget = null;

        // Load Settings
        this._settings = new Gio.Settings({ schema: DOCK_SETTINGS_SCHEMA });
        dockicon_size = this._settings.get_int(DOCK_SIZE_KEY);
        this.iconSize = dockicon_size-8;
        hideDock = hideable = this._settings.get_boolean(DOCK_HIDE_KEY);
        autohide_animation_time = this._settings.get_double(DOCK_AUTOHIDE_ANIMATION_TIME_KEY);

        let favorites = AppFavorites.getAppFavorites().getFavorites();
        this._nicons = favorites.length;

        this._spacing = 4;
        this._item_size = dockicon_size;

        this._box = new St.BoxLayout({ name: 'dock',
                                       vertical: false,
                                       clip_to_allocation: true });
        this._box._delegate = this;

        this.actor = new St.Bin({ x_align: St.Align.START, reactive:true, child: this._box });

        this._workId = Main.initializeDeferredWork(this._box, Lang.bind(this, this._redisplay));

        this._tracker = Shell.WindowTracker.get_default();
        this._appSystem = Shell.AppSystem.get_default();

        this._installedChangedId = this._appSystem.connect('installed-changed', Lang.bind(this, this._queueRedisplay));
        this._appFavoritesChangedId = AppFavorites.getAppFavorites().connect('changed', Lang.bind(this, this._queueRedisplay));
        this._appStateChangedId = this._appSystem.connect('app-state-changed', Lang.bind(this, this._queueRedisplay));

        Main.overview.connect('item-drag-begin',
                              Lang.bind(this, this._onDragBegin));
        Main.overview.connect('item-drag-end',
                              Lang.bind(this, this._onDragEnd));
        Main.overview.connect('item-drag-cancelled',
                              Lang.bind(this, this._onDragCancelled));
        Main.overview.connect('window-drag-begin',
                              Lang.bind(this, this._onDragBegin));
        Main.overview.connect('window-drag-cancelled',
                              Lang.bind(this, this._onDragCancelled));
        Main.overview.connect('window-drag-end',
                              Lang.bind(this, this._onDragEnd));

        this._overviewShowingId = Main.overview.connect('showing', Lang.bind(this, function() {
           hideable = false;
           this._showDock();
        }));
        this._overviewHiddenId = Main.overview.connect('hidden', Lang.bind(this, this._reread_hide_key));

        Main.layoutManager.addChrome(this.actor);

        //hidden
        this._leave_event = this.actor.connect('leave-event', Lang.bind(this, this._hideDock));
        this._enter_event = this.actor.connect('enter-event', Lang.bind(this, this._showDock));

        this._settings.connect('changed::'+DOCK_SIZE_KEY, Lang.bind(this, function (){
            if (!this._settings)
                return;

            dockicon_size = this._settings.get_int(DOCK_SIZE_KEY);
            this._redisplay();
        }));

        this._settings.connect('changed::'+DOCK_HIDE_KEY, Lang.bind(this, this._reread_hide_key));

        this._settings.connect('changed::'+DOCK_AUTOHIDE_ANIMATION_TIME_KEY, Lang.bind(this,function (){
            if (!this._settings)
                return;

            autohide_animation_time = this._settings.get_double(DOCK_AUTOHIDE_ANIMATION_TIME_KEY);
        }));

        this._relocate_panel();
        this._hideDock();
    },

    _relocate_panel: function() {
        let monitor = Main.layoutManager.primaryMonitor;
        this._redisplay();

        let iconChildren = this._box.get_children().filter(function(actor) {
            return actor._delegate.child &&
                   actor._delegate.child._delegate &&
                   actor._delegate.child._delegate.icon &&
                   !actor._delegate.animatingOut;
        });

        let firstIcon = iconChildren[0]._delegate.child._delegate.icon;

        let minWidth, natWidth;

        // Enforce the current icon size during the size request if
        // the icon is animating
        [minWidth, natWidth] = iconChildren[0].get_preferred_width(-1);
        let width = (natWidth+this._spacing)*(this._nicons);
        this.actor.set_position (monitor.x+(monitor.width-width)/2,monitor.y+ monitor.height-5);
    },

    _reread_hide_key: function(){
        if (!this._settings)
            return;

        hideable = this._settings.get_boolean(DOCK_HIDE_KEY);
        if (hideable){
                hideDock=false;
                this._hideDock();
        } else {
                hideDock=true;
                this._showDock();
        }
    },
    
    // fuctions hide
    _restoreHideDock: function(){
        hideable = this._settings.get_boolean(DOCK_HIDE_KEY);
    },

    _disableHideDock: function (){
        hideable = false;
    },

    _queueRedisplay: function () {
        let favorites = AppFavorites.getAppFavorites().getFavorites();
        this._nicons = favorites.length;
        this._showEffectAddItem();
        Main.queueDeferredWork(this._workId);
    },

    _createAppItem: function(app) {
        let display = new AppDisplay.AppWellIcon(app,
                                                 { setSizeManually: true,
                                                   showLabel: false },
                                                function(event){
            this.emit('launching');
            this.app.open_new_window(-1);
            Main.overview.hide();
        });
        display._draggable.connect('drag-begin',
                                   Lang.bind(this, function() {
                                       display.actor.opacity = 50;
                                   }));
        display._draggable.connect('drag-end',
                                   Lang.bind(this, function() {
                                       display.actor.opacity = 255;
                                   }));
        display.actor.set_tooltip_text(app.get_name());

        let item = new DashItemContainer();
        item.setChild(display.actor);

        display.icon.setIconSize(this.iconSize);

        return item;
    },

    _adjustIconSize: function() {
        // For the icon size, we only consider children which are "proper"
        // icons (i.e. ignoring drag placeholders) and which are not
        // animating out (which means they will be destroyed at the end of
        // the animation)

        let iconChildren = this._box.get_children().filter(function(actor) {
            return actor._delegate.child &&
                   actor._delegate.child._delegate &&
                   actor._delegate.child._delegate.icon &&
                   !actor._delegate.animatingOut;
        });


        if (iconChildren.length == 0) {
            this._box.add_style_pseudo_class('empty');
            return;
        }

        this._box.remove_style_pseudo_class('empty');

        if (this._maxWidth == -1)
            return;

        return;
    },

    _redisplay: function () {
        let favorites = AppFavorites.getAppFavorites().getFavoriteMap();

        let children = this._box.get_children().filter(function(actor) {
                return actor._delegate.child &&
                       actor._delegate.child._delegate &&
                       actor._delegate.child._delegate.app;
            });
        // Apps currently in the dash
        let oldApps = children.map(function(actor) {
                return actor._delegate.child._delegate.app;
            });
        // Apps supposed to be in the dash
        let newApps = [];

        for (let id in favorites)
            newApps.push(favorites[id]);

        // Figure out the actual changes to the list of items; we iterate
        // over both the list of items currently in the dash and the list
        // of items expected there, and collect additions and removals.
        // Moves are both an addition and a removal, where the order of
        // the operations depends on whether we encounter the position
        // where the item has been added first or the one from where it
        // was removed.
        // There is an assumption that only one item is moved at a given
        // time; when moving several items at once, everything will still
        // end up at the right position, but there might be additional
        // additions/removals (e.g. it might remove all the launchers
        // and add them back in the new order even if a smaller set of
        // additions and removals is possible).
        // If above assumptions turns out to be a problem, we might need
        // to use a more sophisticated algorithm, e.g. Longest Common
        // Subsequence as used by diff.
        let addedItems = [];
        let removedActors = [];

        let newIndex = 0;
        let oldIndex = 0;
        while (newIndex < newApps.length || oldIndex < oldApps.length) {
            // No change at oldIndex/newIndex
            if (oldApps[oldIndex] == newApps[newIndex]) {
                oldIndex++;
                newIndex++;
                continue;
            }

            // App removed at oldIndex
            if (oldApps[oldIndex] &&
                newApps.indexOf(oldApps[oldIndex]) == -1) {
                removedActors.push(children[oldIndex]);
                oldIndex++;
                continue;
            }

            // App added at newIndex
            if (newApps[newIndex] &&
                oldApps.indexOf(newApps[newIndex]) == -1) {
                addedItems.push({ app: newApps[newIndex],
                                  item: this._createAppItem(newApps[newIndex]),
                                  pos: newIndex });
                newIndex++;
                continue;
            }

            // App moved
            let insertHere = newApps[newIndex + 1] &&
                             newApps[newIndex + 1] == oldApps[oldIndex];
            let alreadyRemoved = removedActors.reduce(function(result, actor) {
                let removedApp = actor._delegate.child._delegate.app;
                return result || removedApp == newApps[newIndex];
            }, false);

            if (insertHere || alreadyRemoved) {
                let newItem = this._createAppItem(newApps[newIndex]);
                addedItems.push({ app: newApps[newIndex],
                                  item: newItem,
                                  pos: newIndex + removedActors.length });
                newIndex++;
            } else {
                removedActors.push(children[oldIndex]);
                oldIndex++;
            }
        }

        for (let i = 0; i < addedItems.length; i++)
            this._box.insert_actor(addedItems[i].item.actor,
                                   addedItems[i].pos);

        for (let i = 0; i < removedActors.length; i++) {
            let item = removedActors[i]._delegate;

            // Don't animate item removal when the overview is hidden
            if (Main.overview.visible)
                item.animateOutAndDestroy();
            else
                item.actor.destroy();
        }

        this._adjustIconSize();

        // Skip animations on first run when adding the initial set
        // of items, to avoid all items zooming in at once
        if (!this._shownInitially) {
            this._shownInitially = true;
            return;
        }

        // Don't animate item addition when the overview is hidden
        if (!Main.overview.visible)
            return;

        for (let i = 0; i < addedItems.length; i++)
            addedItems[i].item.animateIn();
    },

    handleDragOver : function(source, actor, x, y, time) {
        let app = null;
        if (source instanceof AppDisplay.AppWellIcon)
            app = this._appSystem.lookup_app(source.getId());
        else if (source.metaWindow)
            app = this._tracker.get_window_app(source.metaWindow);

        // Don't allow favoriting of transient apps
        if (app == null || app.is_window_backed())
            return DND.DragMotionResult.NO_DROP;

        let favorites = AppFavorites.getAppFavorites().getFavorites();
        let numFavorites = favorites.length;

        let favPos = favorites.indexOf(app);

        let children = this._box.get_children();
        let numChildren = children.length;
        let boxWidth = this._box.width;

        // Keep the placeholder out of the index calculation; assuming that
        // the remove target has the same size as "normal" items, we don't
        // need to do the same adjustment there.
        if (this._dragPlaceholder) {
            boxWidth -= this._dragPlaceholder.actor.width;
            numChildren--;
        }

        let pos = Math.round(x * numChildren / boxWidth);

        if (pos != this._dragPlaceholderPos && pos <= numFavorites) {
            if (this._animatingPlaceholdersCount > 0) {
                let appChildren = children.filter(function(actor) {
                    return actor._delegate &&
                           actor._delegate.child &&
                           actor._delegate.child._delegate &&
                           actor._delegate.child._delegate.app;
                });
                this._dragPlaceholderPos = children.indexOf(appChildren[pos]);
            } else {
                this._dragPlaceholderPos = pos;
            }

            // Don't allow positioning before or after self
            if (favPos != -1 && (pos == favPos || pos == favPos + 1)) {
                if (this._dragPlaceholder) {
                    this._dragPlaceholder.animateOutAndDestroy();
                    this._animatingPlaceholdersCount++;
                    this._dragPlaceholder.actor.connect('destroy',
                        Lang.bind(this, function() {
                            this._animatingPlaceholdersCount--;
                        }));
                }
                this._dragPlaceholder = null;

                return DND.DragMotionResult.CONTINUE;
            }

            // If the placeholder already exists, we just move
            // it, but if we are adding it, expand its size in
            // an animation
            let fadeIn;
            if (this._dragPlaceholder) {
                this._dragPlaceholder.actor.destroy();
                fadeIn = false;
            } else {
                fadeIn = true;
            }

            this._dragPlaceholder = new DragPlaceholderItem();
            this._dragPlaceholder.child.set_width (this.iconSize/2);
            this._dragPlaceholder.child.set_height (this.iconSize);
            this._box.insert_actor(this._dragPlaceholder.actor,
                                   this._dragPlaceholderPos);
            if (fadeIn)
                this._dragPlaceholder.animateIn();
        }

        let srcIsFavorite = (favPos != -1);

        if (srcIsFavorite)
            return DND.DragMotionResult.MOVE_DROP;

        return DND.DragMotionResult.COPY_DROP;
    },

    _hideDock: function() {
         this._item_size = dockicon_size;
         let monitor = Main.layoutManager.primaryMonitor;
         let cornerY = 0;
	 let width = this._nicons*(this._item_size + this._spacing) + 2*this._spacing;
         let height = this._item_size + 4*this._spacing;

         cornerY = monitor.y + monitor.height - this._spacing;

         if (hideable) {
                 Tweener.addTween(this.actor,{
                         y: cornerY,
                         x: monitor.x + (monitor.width - width)/2,
                         height: height,
                         time: autohide_animation_time,
                         transition: 'easeOutQuad'
                       });
                 hideDock=true;
         }
    },

    _showDock: function() {
        this._item_size = dockicon_size;
        let monitor = Main.layoutManager.primaryMonitor;
        let position_y = monitor.y;
        let height = this._item_size + 4*this._spacing;
	let width = this._nicons*(this._item_size + this._spacing) + 2*this._spacing;
        this.actor.set_width(-1);

        position_y=monitor.y + (monitor.height-this._item_size-2*this._spacing);
        Tweener.addTween(this.actor, {
                y: position_y,
                x: monitor.x + (monitor.width - width)/2,
                height: height,
                time: autohide_animation_time,
                transition: 'easeOutQuad'
        });
        hideDock=false;
    },

    _initShowDock : function() {
        this._showDock();
    },

    _showEffectAddItem : function() {
        let monitor = Main.layoutManager.primaryMonitor;
        let height = this._item_size + 4*this._spacing;
	let width = this._nicons*(this._item_size + this._spacing) + 2*this._spacing;

        Tweener.addTween(this.actor, {
                x: monitor.x + (monitor.width-width)/2,
                height: height,
                time: autohide_animation_time,
                transition: 'easeOutQuad'
        });
    },
    destroy: function() {
        if (this._installedChangedId) {
            this._appSystem.disconnect(this._installedChangedId);
            this._installedChangedId = 0;
        }

        if (this._appFavoritesChangedId) {
            AppFavorites.getAppFavorites().disconnect(this._appFavoritesChangedId);
            this._appFavoritesChangedId = 0;
        }

        if (this._appStateChangedId) {
            this._appSystem.disconnect(this._appStateChangedId);
            this._appStateChangedId = 0;
        }

        if (this._overviewShowingId) {
            Main.overview.disconnect(this._overviewShowingId);
            this._overviewShowingId = 0;
        }

        if (this._overviewHiddenId) {
            Main.overview.disconnect(this._overviewHiddenId);
            this._overviewHiddenId = 0;
        }

        this.actor.destroy();

        // Break reference cycles
        this._settings = null;
        this._appSystem = null;
        this._tracker = null;
    },
};

Signals.addSignalMethods(BottomDash.prototype);

let bdock = null;

function init() {
}

function enable() {
    bdock = new BottomDash();
}

function disable() {
    bdock.destroy();
    bdock = null;
}
