// Copyright (C) 2012 Calculate Ltd
// Licence: GPLv2+

const Clutter = imports.gi.Clutter;
const Lang = imports.lang;
const St = imports.gi.St;
const Shell = imports.gi.Shell;
const DateMenu = imports.ui.dateMenu;
const Calendar = imports.ui.calendar;
const Gio = imports.gi.Gio;
const Gdk = imports.gi.Gdk;

const Main = imports.ui.main;
const Mainloop = imports.mainloop;
const PopupMenu = imports.ui.popupMenu;
const userMenuButton = imports.ui.main.panel._statusArea['userMenu'];

const CHANGE_WS_TIME=300;  //time wait before change workspace

function WindowMover(realWindow) {
    this._init(realWindow);
}

WindowMover.prototype = {
    _init : function() {
        this._workspaces = [];
        this._nWorkspacesChanged();
    },

    destroy: function() {
        for(w = 0; w < this._workspaces.length;w++) {
            let workspace = this._workspaces[w];
            workspace.disconnect(workspace._windowPluginAddedId);
            workspace.disconnect(workspace._windowPluginRemovedId);
        }
    },

    _windowRemoved : function(workspace, window) {
        //Main.notify("Window removed");

    },

    _windowAdded : function(workspace, wnd) {
        let win = wnd.get_compositor_private();

        if (!win) {
            // Newly-created windows are added to a workspace before
            // the compositor finds out about them...
            Mainloop.idle_add(Lang.bind(this,
                                        function () {
                                            if (wnd.get_compositor_private())
                                                this._windowAdded(workspace,wnd);
                                            return false;
                                        }));
            return;
        }
        let positionId = win.connect('position-changed', Lang.bind(this,
                        this._changeWorkspace))
        win.connect('destroy', Lang.bind(this,function() {
            win.disconnect(positionId);
        }));
    },

    _nWorkspacesChanged : function() {
        let oldNumWorkspaces = this._workspaces.length;
        let newNumWorkspaces = global.screen.n_workspaces;

        if (oldNumWorkspaces == newNumWorkspaces)
            return false;

        let lostWorkspaces = [];
        if (newNumWorkspaces > oldNumWorkspaces) {
            let w;

            // Assume workspaces are only added at the end
            for (w = oldNumWorkspaces; w < newNumWorkspaces; w++)
                this._workspaces[w] = global.screen.get_workspace_by_index(w);

            for (w = oldNumWorkspaces; w < newNumWorkspaces; w++) {
                let workspace = this._workspaces[w];
                workspace._windowPluginAddedId = workspace.connect('window-added',
                                                           Lang.bind(this, this._windowAdded));
                workspace._windowPluginRemovedId = workspace.connect('window-removed',
                                                           Lang.bind(this, this._windowRemoved));
            }

        } else {
            // Assume workspaces are only removed sequentially
            // (e.g. 2,3,4 - not 2,4,7)
            let removedIndex;
            let removedNum = oldNumWorkspaces - newNumWorkspaces;
            for (let w = 0; w < oldNumWorkspaces; w++) {
                let workspace = global.screen.get_workspace_by_index(w);
                if (this._workspaces[w] != workspace) {
                    removedIndex = w;
                    break;
                }
            }
    
            let lostWorkspaces = this._workspaces.splice(removedIndex, removedNum);
            lostWorkspaces.forEach(function(workspace) {
                                       workspace.disconnect(workspace._windowPluginAddedId);
                                       workspace.disconnect(workspace._windowPluginRemovedId);
                                   });
        }
        return false;
    },

    _changeWorkspace : function() {
        [x,y,z] = global.get_pointer();
        let monitor = Main.layoutManager.primaryMonitor;
        if( x > monitor.x + monitor.width - 2) {
            if(!this._timeout) {
                this._timeout = Mainloop.timeout_add(CHANGE_WS_TIME, Lang.bind(this,function() {
                    let active = global.screen.get_active_workspace_index();
                    let ws = global.screen.get_workspace_by_index(active+1)
                    if(ws) {
                        let display = Gdk.Display.get_default();
                        let deviceManager = display.get_device_manager();
                        let pointer = deviceManager.get_client_pointer();
                        let [screen, pointerX, pointerY] = pointer.get_position();

                        pointer.warp(screen, monitor.x + 5, pointerY);
                        ws.activate(global.get_current_time());
                    }
                    this._timeout = null;
                }));
            }
        }
        else if( x < monitor.x+1) {
            if(!this._timeout) {
                this._timeout = Mainloop.timeout_add(CHANGE_WS_TIME, Lang.bind(this,function() {
                    let active = global.screen.get_active_workspace_index();
                    let ws = global.screen.get_workspace_by_index(active-1)
                    if(ws) {
                        let display = Gdk.Display.get_default();
                        let deviceManager = display.get_device_manager();
                        let pointer = deviceManager.get_client_pointer();
                        let [screen, pointerX, pointerY] = pointer.get_position();

                        pointer.warp(screen, monitor.x+monitor.width-5, pointerY);
                        ws.activate(global.get_current_time());
                    }
                    this._timeout = null;
                }));
            }
        }
        else {
            if(this._timeout) {
                Mainloop.source_remove(this._timeout);
                this._timeout = null;
            }
        }
    }
}

let window_mover;

function init() {
}

function enable() {
    window_mover = new WindowMover();
}

function disable() {
    window_mover.destroy();
}
